# @Created Date: 2020-10-26 01:00:07 pm
# @Filename: warnings.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-10-26 01:00:42 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University


class MultiWrittenWarning(UserWarning):
    pass


class WithoutCifKeyWarning(UserWarning):
    pass

class PISAErrorWarning(UserWarning):
    pass


class MultipleConformersWarning(UserWarning):
    pass
